#!/usr/bin/perl

# Program:  iBatis beans sources and config files generator
# Version:  0.1.9
#
# Author:   alxeg@yandex.ru (Alex Egorov)
#
# Copyright (c) 2003-2004 Alex Egorov. All rights reserved.  This
# script is free software; you may redistribute it under the same terms
# as Perl itself.

use Cwd;

$/=undef; # read whole stream at once, ignoring \n

if (scalar(@ARGV) ==0 ) {
    printUsage();
  exit;
}

our ($packageName, $force, $inputFile, $beansout, $mapsout, $beansinc, $mapsinc);

foreach (@ARGV) {
  if (/^-force/) {
    $force = "true";
  }
  if (/^-input/) {
    if (/^-input=([\w\d\.]+)/) {
      $inputFile = $1;
    }
    else {
      die "No valid input file specified";
    }
  }
  if (/^-beansout/) {
    if (/^-beansout=(.+)/) {
      $beansout = "$1";
      if ( $beansout !~ /[\\\/]$/) {
        $beansout .= "\/";
      }
    }
    else {
      die "No valid java-beans output folder specified";
    }
  }
  if (/^-mapsout/) {
    if (/^-mapsout=(.+)/) {
      $mapsout = "$1";
      if ( $mapsout !~ /[\\\/]$/ ) {
        $mapsout .= "\/";
      }
    }
    else {
      die "No valid sql-maps output folder specified";
    }
  }

  if (/^-beansinc/) {
    if (/^-beansinc=([\w\d]+)/) {
      $beansinc = "$1";
    }
    else {
      die "No valid include file extension specified. ";
    }
  }

  if (/^-mapsinc/) {
    if (/^-mapsinc=([\w\d]+)/) {
      $mapsinc = "$1";
    }
    else {
      die "No valid include file extension specified. ";
    }
  }

  if (/^-package/) {
    if (/^-package=(\w+(\.\w+)*)/) {
      $packageName = $1;
    }
    else {
      die "Wrong package name specified '$1'";
    }
  }
}
if (not $beansout ) {
    $beansout = ".\/";
}
else {
    die "\"$beansout\" folder does not exist" unless (-d $beansout);
}
if (not $mapsout ) {
    $mapsout = ".\/";
}
else {
    die "\"$mapsout\" folder does not exist" unless (-d $mapsout);
}

mktree(split(/\./, $packageName));

our %sqlMapping = readproperties("mapping.properties");

die "No input file specified" unless ($inputFile);
die "File $inputFile doesn't exist" unless (-e $inputFile);

open INPUTFILE, "<$inputFile";

my $text=<INPUTFILE>;

my %tables;

#while ($text =~ /CREATE\s+TABLE\s+(\w+)\s*\((.*?)\);/igs) {
while ($text =~ /CREATE\s+TABLE\s+(\w+)\s*\((.*?)\)[^\)]*;/igs) {
  my $table_name=$1;
  debug("\ntable \"$table_name\"\n");

  (my $fields = $2 ) =~ s/(\n|\s{2,})/ /igs; # remove line breaks and more then 2 spaces
  $fields =~ s/,(?=[\s*\w*,]*\))/;/igs;     # replace ',' with ';' inside parentheses

  foreach my $field ( split(',',$fields)) {
    $field =~ s/^\s+//igs; # remove leading spaces

    #debug(">>>>> $field\n");
    if ($field =~ /^(PRIMARY|UNIQUE|KEY|INDEX|CONSTRAINT)/) {
      if ($field =~ /PRIMARY\s+KEY\s*\(\s*(.*)\s*\)/i) {
        debug("\tprimary key = \"$1\"\n");
        foreach my $key (split(/\s*;\s*/,$1)) {
          %{${$tables{$table_name}}{"ColKeys"}}->{$key}="pk";
        }
      }
      #                                  (   $1       )                $2       (   $3       )
      elsif ($field =~ /FOREIGN\s+KEY\s*\(\s*(\w+)\s*\)\s*REFERENCES\s+(\w+)\s*\(\s*(\w+)\s*\)/i) {
        debug("\tforeign key = field \"$1\" from table \"$2\" on field \"$3\"\n");
        if (exists(%{${$tables{$table_name}}{"ColKeys"}}->{$1})) {
          %{${$tables{$2}}{"ColDepends"}}->{$table_name}->{"field"}=$3;
          %{${$tables{$2}}{"ColDepends"}}->{$table_name}->{"dependField"}=$1;
        }
        else {
          %{${$tables{$table_name}}{"ColDepends"}}->{$2}->{"field"}=$1;
          %{${$tables{$table_name}}{"ColDepends"}}->{$2}->{"dependField"}=$3;
        }
      }
      elsif ($field =~ /^CONSTRAINT\s+(\w+)\s+UNIQUE\s*\((.*)\)/i) {
        # not supported
        debug("\tunique constraint \"$1\" on \"$2\" \n");
      }
    }
    else {
      $field =~ /\s*(\w+)\s+(\w+)\W*/i;
      my $fieldName = $1; my $fieldType = $2;
      debug("\tfield = \"$fieldName\" of type \"$fieldType\"\n");
      %{${$tables{$table_name}}{"ColFields"}}->{$fieldName}=$fieldType;
      if ($field =~ /PRIMARY/i) {
          debug("\tprimary key = \"$fieldName\"\n");
          %{${$tables{$table_name}}{"ColKeys"}}->{$fieldName}="pk";
      }
      elsif ($field =~ /REFERENCES\s+(\w+)\s*\(\w+\)/) {
        debug("\tforeign key = field \"$fieldName\" from table \"$1\" on field \"$2\"\n");
        if (exists(%{${$tables{$table_name}}{"ColKeys"}}->{$fieldName})) {
          %{${$tables{$1}}{"ColDepends"}}->{$table_name}->{"field"}=$2;
          %{${$tables{$1}}{"ColDepends"}}->{$table_name}->{"dependField"}=$fieldName;
        }
        else {
          %{${$tables{$table_name}}{"ColDepends"}}->{$1}->{"field"}=$fieldName;
          %{${$tables{$table_name}}{"ColDepends"}}->{$1}->{"dependField"}=$2;
        }
      }
    }
  }
}

#                               $1                         $2                       (   $3       )                $4       ($5         )
while ($text=~ /ALTER\s+TABLE\s+(\w+)\s+ADD\s+CONSTRAINT\s+(\w+)\s+FOREIGN\s+KEY\s*\(\s*(\w+)\s*\)\s*REFERENCES\s+(\w+)\s*\(([\w\s,]+)\)[^;]*;/igs) {
  debug("for table $1 foreign key = field $3 from table $4 on field $5\n");
  if (exists(%{${$tables{$1}}{"ColKeys"}}->{$3})) {
    %{${$tables{$4}}{"ColDepends"}}->{$1}->{"field"}=$5;
    %{${$tables{$4}}{"ColDepends"}}->{$1}->{"dependField"}=$3;
  }
  else {
    %{${$tables{$1}}{"ColDepends"}}->{$4}->{"field"}=$3;
    %{${$tables{$1}}{"ColDepends"}}->{$4}->{"dependField"}=$5;
  }
}


# Main generation loop
foreach my $tbl (keys %tables) {
  generateBeanFile(
      $tbl,
      \%{${$tables{$tbl}}{"ColFields"}},
      \%{${$tables{$tbl}}{"ColKeys"}},
      \%{${$tables{$tbl}}{"ColDepends"}} );
  generateXmlFile(
      $tbl,
      \%{${$tables{$tbl}}{"ColFields"}},
      \%{${$tables{$tbl}}{"ColKeys"}},
      \%{${$tables{$tbl}}{"ColDepends"}} );
}

sub generateXmlFile($;\%;\%;\%) {
  my ($tableName, $fields, $prKeys, $frKeys) = @_;
  my $tableClassName = genTableName($tableName);

  die "File $mapsout$tableClassName.xml already exists try -force key to recreate all\n" if (-e "$mapsout$tableClassName.xml" && $force ne "true");

  open XMLFILE, ">$mapsout$tableClassName.xml";

  if ($packageName) {
    $packagedClassName = $packageName.".".$tableClassName;
  }
  else  {
    $packagedClassName = $tableClassName;
  }

  print XMLFILE "<?xml version='1.0'?>\n\n";
  print XMLFILE "<!DOCTYPE sql-map PUBLIC \"-//iBATIS.com//DTD SQL Map 1.0//EN\"\n\t\"http://www.ibatis.com/dtd/sql-map.dtd\">\n";

  print XMLFILE "\n<!-- WARNING: This is an autogenerated file -->\n";

  print XMLFILE "\n<sql-map name=\"$tableClassName\">\n";

  my @conditions = ();
  my @setFields = ();
  my @insertFields = ();
  my @insertValues = ();

  my $lowname;
  $lowname = "\L$tableClassName";
  print XMLFILE "\n\t<cache-model name=\"$lowname-cache\" implementation =\"MEMORY\">\n";
  print XMLFILE "\t\t<flush-interval hours=\"24\"/>\n";
  print XMLFILE "\t\t<flush-on-execute statement=\"insert$tableClassName\"/>\n";
  print XMLFILE "\t\t<flush-on-execute statement=\"update$tableClassName\"/>\n";
  print XMLFILE "\t\t<flush-on-execute statement=\"delete$tableClassName\"/>\n";
  print XMLFILE "\t\t<cache-property name=\"reference-type\" value=\"WEAK\" />\n";
  print XMLFILE "\t</cache-model>\n";

  if (scalar(keys(%$fields))!=0) {
    print XMLFILE "\n\t<result-map class=\"$packagedClassName\" name=\"total-\L$tableClassName-result\" >\n";
    foreach my $field (keys(%$fields)) {
      my $fieldName = genFieldName($field);
      print XMLFILE "\t\t<property name=\"$fieldName\" column=\"$field\" />\n";
      if (exists(%$prKeys->{$field})) {
        push @conditions, " $field = #$fieldName# ";
      }
      else {
        push @setFields, " $field = #$fieldName# ";
      }
      push @insertFields, "$field";
      push @insertValues, "#$fieldName#";
    }
    print XMLFILE "\t</result-map>\n";
  }

  my $condition = join(" and ", @conditions);
  my $setFields = join(", ", @setFields);
  print XMLFILE "\n\t<mapped-statement name=\"get$tableClassName\" result-class=\"$packagedClassName\" parameter-class=\"$packagedClassName\" result-map=\"total-\L$tableClassName-result\" >\n";
  print XMLFILE "\t\t<![CDATA[\n";
  print XMLFILE "\t\t\tselect * from $tableName\n";
  print XMLFILE "\t\t\twhere $condition\n";
  print XMLFILE "\t\t]]>\n";
  print XMLFILE "\t</mapped-statement>\n";

  print XMLFILE "\n\t<mapped-statement name=\"update$tableClassName\" parameter-class=\"$packagedClassName\" >\n";
  print XMLFILE "\t\t<![CDATA[\n";
  print XMLFILE "\t\t\tupdate $tableName\n";
  print XMLFILE "\t\t\tset $setFields\n";
  print XMLFILE "\t\t\twhere $condition\n";
  print XMLFILE "\t\t]]>\n";
  print XMLFILE "\t</mapped-statement>\n";

  print XMLFILE "\n\t<mapped-statement name=\"insert$tableClassName\" parameter-class=\"$packagedClassName\" >\n";
  print XMLFILE "\t\t<![CDATA[\n";
  print XMLFILE "\t\t\tinsert into $tableName(", join(", ", @insertFields),")\n";
  print XMLFILE "\t\t\tvalues(", join(", ", @insertValues),")\n";
  print XMLFILE "\t\t]]>\n";
  print XMLFILE "\t</mapped-statement>\n";

  print XMLFILE "\n\t<mapped-statement name=\"delete$tableClassName\" parameter-class=\"$packagedClassName\" >\n";
  print XMLFILE "\t\t<![CDATA[\n";
  print XMLFILE "\t\t\tdelete from $tableName\n";
  print XMLFILE "\t\t\twhere $condition\n";
  print XMLFILE "\t\t]]>\n";
  print XMLFILE "\t</mapped-statement>\n";


  if ($mapsinc) {
    open INCFILE, "<$mapsout$tableClassName.xml.$mapsinc";
    my $inctext=<INCFILE>;
    print XMLFILE "\n$inctext\n";
    close (INCFILE);
  }

  #print XMLFILE "";

  print XMLFILE "\n</sql-map>";

  close(XMLFILE);
}

sub generateBeanFile($;\%;\%;\%) {
  my ($tableName, $fields, $prKeys, $frKeys) = @_;
  my $tableClassName = genTableName($tableName);
  my $packageDir;
  if ($packageName) {
    ($packageDir = $packageName) =~ s/\./\//ig;
    $packageDir .= "\/";
  }

  #print "Generating $beansout$packageDir$tableClassName.java...\n";
  die "File $beansout$packageDir$tableClassName.java already exists try -force key to recreate all\n" if (-e "$beansout$packageDir$tableClassName.java" && $force ne "true");

  open JAVAFILE, ">$beansout$packageDir$tableClassName.java" || die "Can't open file $beansout$packageDir$tableClassName.java for writing";

  #print "1\n";

  if ($packageName) {
    print JAVAFILE "package $packageName;\n";
  }
  print JAVAFILE "\n// WARNING: This is an autogenerated file\n";
  print JAVAFILE "\npublic class $tableClassName {\n";

  print JAVAFILE "\n\t// Constructors\n";
  print JAVAFILE "\tpublic $tableClassName() {\n";
  print JAVAFILE "\t}\n";

  if (scalar keys %$prKeys!=0) {
    my $constructorBody;
    my @parameters = ();
    foreach my $key (keys %$prKeys) {
      my $typeName = getJavaType($fields->{$key});
      my $fieldName = genFieldName($key);
      push @parameters, "$typeName _$fieldName";
      $constructorBody = $constructorBody . "\t\tthis.$fieldName = _$fieldName;\n";
    }
    print JAVAFILE "\tpublic $tableClassName(", join (', ', @parameters), ") {\n";
    print JAVAFILE $constructorBody;
    print JAVAFILE "\t}\n";
  }

  print JAVAFILE "\n\t// Fields";
  foreach my $field (keys(%$fields)) {
    my $type=$fields->{$field};
    print JAVAFILE genField($field, $type, "true");
    #if (exists(%$prKeys->{$field})) {
    # primary key
    #}
  }

  if (scalar(keys(%$frKeys))!=0) {
    print JAVAFILE "\n\t// ATTENTION: This method(s) just relies on database structure  \n";
    print JAVAFILE   "\t//            and may not correspond with logical structure of appication\n";
  }
  foreach my $dep (keys(%$frKeys)) {
    my $fldName = %{$frKeys}->{$dep}->{"field"};
    $fldName = genFieldName($fldName);
    my $className = genTableName($dep);
    print JAVAFILE "\tpublic $className get$className() {\n";
    print JAVAFILE "\t\t// TODO: Implement getting $className by $fldName\n";
    print JAVAFILE "\t\treturn null;\n";
    print JAVAFILE "\t}\n";

    #print JAVAFILE genField($dep, genTableName($dep), "false");
  }

  if ($beansinc) {
    open INCFILE, "<$beansout$packageDir$tableClassName.java.$beansinc";
    my $inctext=<INCFILE>;
    print JAVAFILE "\n$inctext\n";
    close (INCFILE);
  }

  print JAVAFILE "\n}\n";

  close(JAVAFILE);
}

sub genTableName($) {
  my $name = $_[0];
  $name =~ s/(\w+)/\L$1/ig;
  $name =~ s/_(\w)/\U$1/ig;
  $name =~ s/^(\w)/\U$1/ig;
  return $name;
}

sub genFieldName($) {
  my $name = $_[0];
  $name =~ s/(\w+)/\L$1/ig;
  $name =~ s/_(\w)/\U$1/ig;
  return $name;
}

sub genField($;$;$) {
  my ($field, $type, $lookupJavaType) = @_;

  my $name = genFieldName($field);
  my $javaType = $type;
  if ($lookupJavaType =~ /true/) {
    $javaType =  getJavaType($type);
  }
  my $result = "\n\tprivate $javaType $name;\n";
  (my $capitalName = $name) =~ s/^(\w)/\U$1/ig;
  $result = $result . "\tpublic $javaType get$capitalName() {\n";
  $result = $result . "\t\treturn $name;\n";
  $result = $result . "\t}\n";
  $result = $result . "\tpublic void set$capitalName($javaType _$name) {\n";
  $result = $result . "\t\tthis.$name = _$name;\n";
  $result = $result . "\t}\n";
  return $result;
}

sub getJavaType {
  my $upperType = "\U$_[0]";
  debug("finding sql type for \"$upperType\"\n");
  if (!exists($sqlMapping{$upperType})) {
    die "Can't find java mapping for type $_[0]";
  }
  return $sqlMapping{$upperType};
}


sub readproperties($) {
  my ($propfile) = @_;
  my (%props);    # Return value: a hash of property values.
  my ($ln,$name,$value,@nv);
  debug("reading properties file \"$propfile\"\n");
  open(PROPFILE,"$propfile") or die "Unable to open $propfile";

  foreach (split /\n/, <PROPFILE>)  {
    chomp;
    s/#(.*$)//;
    if ($1 ne "") { debug("# $1\n"); }
    @nv = split /\s*=/,$_,2;
    $value = pop @nv;
    $value =~ s/^\s*=?\s*//;
    $value =~ s/\s*$//;
    $name = pop @nv;
    $name =~ s/^\s*//;
    $name =~ s/\s*$//;
    if ($name ne "") {
      $props{$name} = $value;
      debug("$name=\"$value\"\n");
    }
  }
  close(PROPFILE);
  debug("end reading properties file");
  return %props;
}

sub mktree {
  my @path = @_;
  my $curDir = getcwd();
  chdir $beansout;
  foreach (@path) {
    if (-e) {
      if (-d) { chdir $_; }
      else {  die "$_ exists and is not directory"; }
    }
    else {
      mkdir $_;
      chdir $_;
    }
  }
  chdir $curDir;
}

sub printUsage {
  print "\niBatis java-beans sources and sql-map config files generator.\n";
  print "Version 0.1.9\n\n";
  print "http://alxeg.narod.ru/ibatis/index.html\n\n";
  print "USAGE: sql2iBatis.pl -input=<sqlfile> [OPTIONS]\n";
  print "\t-input=<sqlfile>\tsql-file to parse\n";
  print "OPTIONS are:\n";
  print "\t-package=<packageName>\toptional package name (domain.company.project)\n";
  print "\t-force\t\t\tforces files regeneration\n";
  print "\t-beansout\t\tjava-beans output folder\n";
  print "\t-mapsout\t\tsql-maps output folder\n";
  print "\t-mapsinc\t\tExtension of file containing user-defined mapped statements\n";
  print "\t-beansinc\t\tExtension of file containing user-defined java-beans code\n";
  print "\nNOTE:\tFile with an extension specified using the parameter -beansinc (e.g. 'UserProperties.xml.inc')\n";
  print "\tmust be located in sql-maps output folder. It's content will be inserted right after the generated\n";
  print "\tmapped-statements in the destination file.\n";

  #print "\t\twill be added after generated mapping";

}

sub debug {
  #uncomment next line for debug messages
  #print @_;
}
